// jshint maxerr:1000
/*
 * Copyright (c) 2017, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and themofollowing disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* ===============================================================================================================
   ===============================================================================================================
                                                     BASIC TAB TOOL 
   ===============================================================================================================
   =============================================================================================================== */

(function () {
    // ================================ START: CONSTRUCTORS  ====================================================  
    // Constructor call for mmWaveBasic Obj
    function mmWaveBasic() {
        if (!(this instanceof mmWaveBasic))
            return new mmWaveBasic();
        this.init();
    };
    
    // Init call used in constructor for the mmWaveBasic Obj
    function init() {
        this.Input = Basic_Globals.BackendInputs;
        this.Intermediate = Basic_Globals.BackendIntermediates;
        this.Output = Basic_Globals.BackendOutputs;
       
    };
    // ================================ END: CONSTRUCTORS  ==================================================================
        // ================================ START: UPDATE FUNCTIONS  ============================================================ 
    /**
     * This is a helper function used by updateInput() to calculate all the Input Object's values from any changes in value.
     * @param {*} basicToolObj - "this", the self that this function is updating
     * @param {*} mode - boolean for mode of operation,
     *                   0 : standard update, 
     *                   1 : iterative/convergance mode to solve for an adc_valid_start time & idle time
     * @returns - the updated object
     */
    var updateOutputCompHelper = function(basicObject, mode){
        // Grab the self
        var Input           = basicObject.Input;
        var Intermediate    = basicObject.Intermediate;
        var Output          = basicObject.Output;
        
        // Handle the Device and EVM Specific Parameters here
        Output.num_rx               = get_num_rx(Input.device);
        Output.num_tx               = get_num_tx(Input.device);
        Output.tx_gain              = get_tx_gain(Input.device, Input.evm);
        Output.rx_gain              = get_rx_gain(Input.device, Input.evm);
        Output.maximum_bandwidth    = get_maximum_bandwidth(Input.device, Input.frequency_range);
        Output.tx_power             = get_tx_power(Input.device, Input.evm)

        // Percolate changes through all the values:
        Output.bursts_in_frame = get_bursts_in_frame(Input.device);
		Intermediate.maximum_radar_cube_size = get_maximum_radar_cube_size(Input.device);
		Intermediate.maximum_IF_bandwidth = get_maximum_IF_bandwidth(Input.device);
		Intermediate.maximum_sampling_frequency = get_maximum_sampling_frequency(Input.device);
		Output.maximum_bandwidth = get_maximum_bandwidth(Input.device, Input.frequency_range);
		Output.starting_frequency = get_starting_frequency(Input.device, Input.frequency_range);
		Intermediate.maximum_velocity = get_maximum_velocity(Input.maximum_velocity_kmph);
		Intermediate.velocity_resolution = get_velocity_resolution(Input.velocity_resolution_kmph);
		Intermediate.valid_sweep_bandwidth = get_valid_sweep_bandwidth(Basic_Globals.Constants.lightSpeed, Input.range_resolution);
        if (mode) {
            if (Intermediate.idle_time < Output.idle_time_minimum) {
                Intermediate.idle_time = Output.idle_time_minimum;
            }
            if (Intermediate.adc_valid_start_time != Output.adc_valid_start_time_2) {
                Intermediate.adc_valid_start_time = Output.adc_valid_start_time_2;
            }
        }
		else {
            Intermediate.idle_time = get_idle_time(Intermediate.valid_sweep_bandwidth);
            Intermediate.adc_valid_start_time = get_adc_valid_start_time();
        }
        Intermediate.excess_ramping_time = get_excess_ramping_time();
		Intermediate.periodicity = get_periodicity(Input.measurement_rate);
		Intermediate.ambient_temperature = get_ambient_temperature();
		Intermediate.noise_figure = get_noise_figure();
        Intermediate.num_virtual_rx = get_num_virtual_rx(Output.num_rx, Output.num_tx);   
		Intermediate.non_coherent_combining_loss = get_non_coherent_combining_loss(Intermediate.num_virtual_rx);
		Intermediate.rcs_value = get_rcs_value(Input.typical_detected_object);
		Intermediate.combined_factor_in_dB = get_combined_factor_in_dB(Output.tx_power, Output.tx_gain, Output.rx_gain, Intermediate.non_coherent_combining_loss,
		                                                    Input.detection_loss, Input.system_loss, Input.implementation_margin, Input.detection_SNR, Intermediate.noise_figure);
		Intermediate.combined_factor_linear = get_combined_factor_linear(Intermediate.combined_factor_in_dB);
		Intermediate.inter_chirp_time = get_inter_chirp_time(Intermediate.idle_time, Intermediate.adc_valid_start_time, Intermediate.excess_ramping_time);
		Intermediate.aux_comp_coeff_a = get_aux_comp_coeff_a(Output.starting_frequency, Intermediate.valid_sweep_bandwidth);
		Intermediate.aux_comp_coeff_b = get_aux_comp_coeff_b(Intermediate.valid_sweep_bandwidth, Intermediate.adc_valid_start_time,
		                                                    Output.starting_frequency, Intermediate.inter_chirp_time, Basic_Globals.Constants.lightSpeed, Intermediate.maximum_velocity, Output.num_tx);
		Intermediate.chirp_time_temp = get_chirp_time_temp(Intermediate.aux_comp_coeff_b, Intermediate.aux_comp_coeff_a); 
		Intermediate.ramp_slope_init = get_ramp_slope_init(Intermediate.valid_sweep_bandwidth, Intermediate.chirp_time_temp);
		Output.ramp_slope_parameter = get_ramp_slope_parameter(Intermediate.ramp_slope_init);
		Output.ramp_slope = get_ramp_slope(Output.ramp_slope_parameter);
		Intermediate.aux_comp_T1 = get_aux_comp_T1(Output.ramp_slope);
		Output.maximum_beat_frequency = get_maximum_beat_frequency(Output.ramp_slope, Input.maximum_detectable_range, Basic_Globals.Constants.lightSpeed);
		Output.sampling_frequency = get_sampling_frequency(Output.maximum_beat_frequency, Input.device);
		Output.number_of_samples_per_chirp = get_number_of_samples_per_chirp(Intermediate.chirp_time_temp, Output.sampling_frequency, Input.device);
		Output.total_sweep_bandwidth = get_total_sweep_bandwidth(Output.ramp_slope, Intermediate.adc_valid_start_time, 
                                                            Output.number_of_samples_per_chirp, Output.sampling_frequency, Intermediate.excess_ramping_time);
		Output.idle_time_minimum = get_idle_time_minimum(Input.device, Output.total_sweep_bandwidth);
        Output.ramp_end_time = get_ramp_end_time(Intermediate.adc_valid_start_time, Output.number_of_samples_per_chirp, Output.sampling_frequency, Intermediate.excess_ramping_time);
        Intermediate.carrier_frequency = get_carrier_frequency(Output.starting_frequency, Output.ramp_slope, Intermediate.adc_valid_start_time,
                                                            Output.number_of_samples_per_chirp, Output.ramp_end_time);
		Intermediate.aux_comp_T2 = get_aux_comp_T2(Output.sampling_frequency);
        Output.adc_valid_start_time_2 = get_adc_valid_start_time_2(Intermediate.aux_comp_T1, Intermediate.aux_comp_T2);
        Intermediate.lambda = get_lambda(Basic_Globals.Constants.lightSpeed, Intermediate.carrier_frequency);
        
        Output.frame_periodicity = get_frame_periodicity(Input.measurement_rate);    
        Intermediate.max_burst_periodicity = get_max_burst_periodicity(Output.frame_periodicity, Output.bursts_in_frame);
        
        Intermediate.max_chirp_repetition_period = get_max_chirp_repetition_period(Intermediate.lambda, Intermediate.maximum_velocity);
        Output.chirp_repetition_period = get_chirp_repetition_period(Output.num_tx, Intermediate.idle_time, Output.ramp_end_time);
        Output.min_num_of_chirp_loops = get_min_num_of_chirp_loops(Basic_Globals.Constants.lightSpeed, Intermediate.carrier_frequency,
            Output.chirp_repetition_period, Intermediate.velocity_resolution);
        Output.chirps_in_burst = get_chirps_in_burst(Output.min_num_of_chirp_loops, Output.bursts_in_frame);
        Output.burst_periodicity = get_burst_periodicity(Output.chirp_repetition_period, Intermediate.max_burst_periodicity, Output.chirps_in_burst, Output.num_tx);
        Output.num_range_fft_bins = get_num_range_fft_bins(Output.number_of_samples_per_chirp);
		Output.max_range_for_typical_detectable_object = get_max_range_for_typical_detectable_object(Intermediate.rcs_value, Intermediate.combined_factor_linear, 
                                                            Intermediate.lambda, Intermediate.num_virtual_rx,Intermediate.chirp_time_temp, Output.min_num_of_chirp_loops, 
                                                            Basic_Globals.Constants.cube_4pi, Basic_Globals.Constants.kB, Intermediate.ambient_temperature);
		Output.min_rcs_detectable_at_max_range = get_min_rcs_detectable_at_max_range(Input.maximum_detectable_range, Basic_Globals.Constants.cube_4pi, Basic_Globals.Constants.kB, 
                                                            Intermediate.ambient_temperature, Intermediate.combined_factor_linear, Intermediate.lambda, Intermediate.num_virtual_rx, 
                                                            Intermediate.chirp_time_temp, Output.min_num_of_chirp_loops);
        Output.num_doppler_fft_bins = get_num_doppler_fft_bins(Output.min_num_of_chirp_loops);
        Output.active_frame_time = get_active_frame_time(Output.min_num_of_chirp_loops, Output.chirp_repetition_period, Output.chirps_in_burst, Output.bursts_in_frame);
        Output.range_inter_bin_resolution = get_range_inter_bin_resolution(Input.range_resolution, Output.number_of_samples_per_chirp, Output.num_range_fft_bins);
        Output.velocity_inter_bin_resolution = get_velocity_inter_bin_resolution(Intermediate.velocity_resolution, Output.min_num_of_chirp_loops, Output.num_doppler_fft_bins);
        Output.radar_cube_size = get_radar_cube_size(Output.num_range_fft_bins, Output.min_num_of_chirp_loops, Intermediate.num_virtual_rx);
        Output.chirp_time = get_chirp_time(Output.ramp_end_time, Output.idle_time_minimum);
    };

    // Performs the iterative convergance loop for idle time (Luke here: why this is like this, I have no idea, and I don't like it)
    var updateOutputComputation = function(basicObject){
        // Grab the self:
        var basic = basicObject;

        // Clear the prior values before movie forward
        for (var key in basic.Intermediate) basic.Intermediate[key] = '';
        for (key in basic.Output){
            if (key === 'num_rx' || key === 'num_tx' || key === 'tx_gain' || key === 'rx_gain' || key === 'maximum_bandwidth' || key === 'tx_power');
            else basic.Output[key] = '';
        }

        // Percolate inputs once
        updateOutputCompHelper(basic,0); // Pass in the input field of the basic Obj
        
        var i = 0;
        // This is very ugly... Why is there a while loop here?
        while (basic.Intermediate.idle_time < basic.Output.idle_time_minimum || basic.Intermediate.adc_valid_start_time != basic.Output.adc_valid_start_time_2) {
            if (i > 10) {
                break;
            }
            //console.log("Idle Time and ADC Start Time min search iteration #" + i);
            i++;
            //console.log("Intermediate Valid ADC Start Time: " + basic.Intermediate.adc_valid_start_time);   
            // Percolate changes through again until the start times above have settled:
            updateOutputCompHelper(basic,1);
        }
        //console.log("Idle Time: " + basic.Intermediate.idle_time);
        //console.log("Valid ADC Start Time: " + basic.Intermediate.adc_valid_start_time);
    }
	
    var updateInput = function(changes) {
        // Grab the self:
        var basic = this;

        // Log the changes that have occurred to the input side of the object:
        for (var k in changes) {
            if (changes.hasOwnProperty(k)) {
                basic.Input[k] = changes[k];  // Copy over the changes into the input object
            }
        }

        updateOutputComputation(basic);
    };
    // ================================ END: UPDATE FUNCTIONS  ============================================================== 
    // ================================ START: GETTER FUNCTIONS  ============================================================    
    // Getter functions for the Device and EVM Specific Parameters
    var get_evm_list = function(device){
        return Object.keys(Basic_Globals.Platforms[device].evms);
    };
    var get_num_rx = function(device){
        return Basic_Globals.Platforms[device].num_rx;
    };
    var get_num_tx = function(device){
        return Basic_Globals.Platforms[device].num_tx;
    };
    var get_rx_gain = function(device, evm){
        return Basic_Globals.Platforms[device].evms[evm].rx_gain;
    };
    var get_tx_gain = function(device, evm){
        return Basic_Globals.Platforms[device].evms[evm].tx_gain;
    };
    var get_allowed_freq_range = function(device){
        var allowed = [];
        Basic_Globals.Platforms[device].freq_range.forEach(list => allowed.push(list[2]));
        return allowed;
    };
    var get_tx_power = function(device, evm){
        return Basic_Globals.Platforms[device].evms[evm].tx_power;
    }
    // Get the device's max radar cube storage size in kB, which is equivalent to 'L3 RAM size' in the data sheet.  
    // https://www.ti.com/lit/an/swra663/swra663.pdf?ts=1655232164682&ref_url=https%253A%252F%252Fwww.google.com%252F
    var get_maximum_radar_cube_size = function(device) {
        return Basic_Globals.Platforms[device].maximum_radar_cube_size;
	};
    // Get the device's max RF bandwidth in MHz, which is 'Max IF Bandwidth' in the datasheet.
    var get_maximum_IF_bandwidth = function(device) {
        return Basic_Globals.Platforms[device].maximum_IF_bandwidth;
	};
    // Get the max complex sampling rate in Msps, which is found in the datasheet.
	var get_maximum_sampling_frequency = function(device) {
        return Basic_Globals.Platforms[device].maximum_sampling_frequency;
	};
	var get_maximum_bandwidth = function(device, frequency_range) {
        var value;
        Basic_Globals.Platforms[device].freq_range.forEach(list => {
            if (list[2] === frequency_range) value = (list[1]-list[0]) * 1000;
        });
        return value;
	};
	var get_starting_frequency = function(device, frequency_range) {
        var value;
        Basic_Globals.Platforms[device].freq_range.forEach(list => {
            if (list[2] === frequency_range) value = list[0]; 
        });
        return value;
	};	
    // Convert the max velocity in km/hr to m/s
	var get_maximum_velocity = function(maximum_velocity_kmph) {
        return maximum_velocity_kmph*1000/3600;  // in m/s
	};
    // Convert the velocity resolution in km/hr to m/s
	var get_velocity_resolution = function(velocity_resolution_kmph) {
        return velocity_resolution_kmph*1000/3600; // m/s
	};
	var get_txPowerCombinedTx = function(noOfTx, baseTxPower) {
        return  Math.floor(baseTxPower+20*Math.log10(noOfTx));
	};
    // Return an idle time in mircoseconds (us) based on the valid sweep bandwidth
	var get_idle_time = function(valid_sweep_bandwidth) {
        if (valid_sweep_bandwidth <= 1000) {
            return 2;   // us
	    } 
        else if (valid_sweep_bandwidth <= 2000) {
            return 5;   // us
		} 
        else if (valid_sweep_bandwidth <= 3000) {
            return 6;   // us
		} 
        else {
            return 7;   // us
		}
	};
	var get_adc_valid_start_time =  function() {
        return 2.8;
	};
	var get_excess_ramping_time = function() {
        return 1;
	};
    // Convert a Hz measurement rate and convert it to ms measurement period
	var get_periodicity = function(measurement_rate) {
        return 1/measurement_rate*1000; // ms
	};
	var get_aux_comp_coeff_a = function(starting_frequency, valid_sweep_bandwidth) {
        return (starting_frequency*1000+valid_sweep_bandwidth/2)*1000000;
	};
	var get_aux_comp_coeff_b = function(valid_sweep_bandwidth, adc_valid_start_time, starting_frequency, inter_chirp_time, lightSpeed, maximum_velocity, num_tx) {
        return (valid_sweep_bandwidth*adc_valid_start_time +
            starting_frequency*1000*inter_chirp_time + 
            inter_chirp_time*valid_sweep_bandwidth/2 -
            lightSpeed/(4*maximum_velocity*num_tx));
    };
	var get_aux_comp_T1 = function(ramp_slope) {
        if (ramp_slope < 50) {
            return 1;
		} else {
            return 2.5;
		}
	};
	var get_aux_comp_T2 = function(sampling_frequency) {
        return Util.Format.toPrecision(3-4.9*(1/(sampling_frequency))+36.5*(1/(sampling_frequency))*(1/(sampling_frequency)), 1);
    };
    // Convert the input ambient temperature in Celcius to Kelvin
	var get_ambient_temperature = function() {
		return 20 + 273.15;
	};
	var get_noise_figure = function() {
		return 14;
	};
    // Get the number of virtual input channels/antennas from the number of Rx and Tx antennas
	var get_num_virtual_rx = function(num_rx, num_tx) {
        return num_rx*num_tx;
	};
	var get_non_coherent_combining_loss = function(num_virtual_rx) {
		if (num_virtual_rx >= 8) {
			return 3;
		} else if (num_virtual_rx == 4) {
			return 2;
		} else if (num_virtual_rx == 2) {
			return 1;
		} else {
			return 0;
		}		
	};
	var get_typical_detected_object_to_rcs = function(typical_detected_object) {
 		if (typical_detected_object == 'Truck')
 		    return 100;
 		if (typical_detected_object == 'Car')
 			return 5;
 		if (typical_detected_object == 'Motorcycle')
 			return 3.2;
 		if (typical_detected_object == 'Adult')
 			return 1;
 		if (typical_detected_object == 'Child')
 		    return 0.5;
 		return undefined;
	};
	// Based on whether a typical detected object or a custom input is used, get the radar cross-section (m^2)
	var get_rcs_value = function(typical_detected_object) {
		if (typical_detected_object && typeof(typical_detected_object) == 'string') {
		    var ans = get_typical_detected_object_to_rcs(typical_detected_object);
		    if (ans !== undefined) return ans;
		}
		var tmp = parseFloat(typical_detected_object);
		if (isNaN(tmp) === true) {
		    return 1;
		}
		return tmp;
	};	
	var get_combined_factor_in_dB = function(tx_power, tx_gain, rx_gain, non_coherent_combining_loss,
	    detection_loss, system_loss, implementation_margin, detection_SNR, noise_figure) {
		return tx_power+tx_gain+rx_gain-non_coherent_combining_loss-
			detection_loss-system_loss-implementation_margin-detection_SNR-noise_figure;
	};	
	var get_combined_factor_linear = function(combined_factor_in_dB) {
		return Math.pow(10, combined_factor_in_dB/10);
	};	
    var get_max_range_for_typical_detectable_object = function(rcs_value, combined_factor_linear,
        lambda, num_virtual_rx, chirp_time_temp, min_num_of_chirp_loops, cube_4pi, kB, ambient_temperature) {
        var maxDetectRange = Math.sqrt(Math.sqrt((0.001*rcs_value*combined_factor_linear*Math.pow(lambda,2)*num_virtual_rx*chirp_time_temp*min_num_of_chirp_loops)/(0.8*cube_4pi*kB*ambient_temperature*1e12)));
        if ((templateObj.$.ti_widget_droplist_device.selectedValue == 'xWRL6432' || templateObj.$.ti_widget_droplist_device.selectedValue == 'xWRL1432') && templateObj.$.ti_widget_droplist_desired_cfg.selectedValue == 'Short Range Default') {
            if (maxDetectRange > 15) maxDetectRange -= 15; // Poor practice, but this value is misleading for low-power devices, so scaling accordingly.
        }
        return maxDetectRange;
    };    
    var get_min_rcs_detectable_at_max_range = function(maximum_detectable_range, cube_4pi, kB,
        ambient_temperature, combined_factor_linear, lambda, num_virtual_rx, chirp_time_temp, min_num_of_chirp_loops) {
        return (0.8*Math.pow(maximum_detectable_range, 4)*
        cube_4pi*kB*ambient_temperature*1e12)/(0.001*combined_factor_linear*lambda*lambda*num_virtual_rx*chirp_time_temp*min_num_of_chirp_loops);
    };	
	var get_valid_sweep_bandwidth = function(lightSpeed, range_resolution) {
		return lightSpeed/(2*range_resolution*0.01*1000000);
	};	
	var get_inter_chirp_time = function(idle_time, adc_valid_start_time, excess_ramping_time) {
		return idle_time+adc_valid_start_time+excess_ramping_time;
	};
	var get_chirp_time_temp = function(aux_comp_coeff_b, aux_comp_coeff_a) {
		return -1000000*aux_comp_coeff_b/aux_comp_coeff_a;
	};	
	var get_ramp_slope_init = function(valid_sweep_bandwidth, chirp_time_temp) {
		return valid_sweep_bandwidth/chirp_time_temp;
	};	
	var get_ramp_slope_parameter = function(ramp_slope_init) {
		return Math.round(((ramp_slope_init*(1<<26)*1000000)/(3600000000*900)) );
	};	
	var get_ramp_slope = function(ramp_slope_parameter) {
		return (ramp_slope_parameter*3600000000*900)/((1<<26)*1000000);
	};
	var get_maximum_beat_frequency = function(ramp_slope, maximum_detectable_range, lightSpeed) {
		return 2000000*ramp_slope*maximum_detectable_range/lightSpeed;
	};
	var get_sampling_frequency = function(maximum_beat_frequency, device) {
		if (device == 'xWRL6432' || device == 'xWRL1432'){ //real only
            return Math.max((2*maximum_beat_frequency)/0.8, 3);
        }
        else{ //complex 1x
            return Math.max((maximum_beat_frequency)/0.8, 3);
        }
	};
	var get_number_of_samples_per_chirp = function(chirp_time_temp, sampling_frequency, device) {
        var samps_per_chirp = 0;
        samps_per_chirp = Math.floor(chirp_time_temp*sampling_frequency)
        /* XWRL6432 SDK 5.1.0.4 supports inputs of powers of 2 up to 10 (2,4,8,16 etc)  */
        if (device == 'xWRL6432' || device == 'xWRL1432'){
            samps_per_chirp = 128;
        }
		return samps_per_chirp;
	};
	var get_total_sweep_bandwidth = function(ramp_slope, adc_valid_start_time, number_of_samples_per_chirp, sampling_frequency, excess_ramping_time) {
	    return ramp_slope*(adc_valid_start_time+number_of_samples_per_chirp/sampling_frequency+excess_ramping_time);
	};
	var get_idle_time_minimum = function(device, total_sweep_bandwidth) {
	    // XWRL6432 has its own idle time constraints for low power operation
        if (device == 'xWRL6432' || device == 'xWRL1432'){
            return 6; // 6us is the safe idle time in all situations
        }

        // In other cases, calculate the idle time based on the total sweep bandwidth
        else {
            if (total_sweep_bandwidth < 1000) {
                return 2;
            } else if (total_sweep_bandwidth < 2000) {
                return 5;
            } else if (total_sweep_bandwidth < 3000) {
                return 6;
            } else {
                return 7;
            }
        }
        
	};
	var get_ramp_end_time = function(adc_valid_start_time, number_of_samples_per_chirp, sampling_frequency, excess_ramping_time) {
	    return adc_valid_start_time+number_of_samples_per_chirp/sampling_frequency+excess_ramping_time;
	};
	var get_carrier_frequency = function(starting_frequency, ramp_slope, adc_valid_start_time, number_of_samples_per_chirp, ramp_end_time) {
	    return (starting_frequency*1000+ramp_slope*adc_valid_start_time+number_of_samples_per_chirp/(2*ramp_end_time*1000))/1000;
	};
	var get_adc_valid_start_time_2 = function(aux_comp_T1, aux_comp_T2) {
	    return aux_comp_T1+aux_comp_T2;
	};
	var get_lambda = function(lightSpeed, carrier_frequency) {
	    return (lightSpeed/carrier_frequency)/1000000;
	};
	var get_max_chirp_repetition_period = function(lambda, maximum_velocity) {
	    return Util.Format.toPrecision((1000*lambda)/(4*maximum_velocity), 0);
	};
	var get_chirp_repetition_period = function(num_tx, idle_time, ramp_end_time) {
            return Util.Format.toPrecision(num_tx*(idle_time+ramp_end_time), 0);
	};
	var get_num_range_fft_bins = function(number_of_samples_per_chirp) {
	    return Math.pow(2,Math.ceil(Math.log2(number_of_samples_per_chirp)));
	};
	var get_min_num_of_chirp_loops = function(lightSpeed, carrier_frequency, chirp_repetition_period, velocity_resolution) {
	    return Math.ceil(lightSpeed/(2*carrier_frequency*chirp_repetition_period*velocity_resolution*1000));
	};
	var get_num_doppler_fft_bins = function(min_num_of_chirp_loops) {
	    return Math.pow(2,Math.ceil(Math.log2(min_num_of_chirp_loops)));
	};
	var get_active_frame_time = function(min_num_of_chirp_loops, chirp_repetition_period) {
	    return min_num_of_chirp_loops*chirp_repetition_period/1000;
	};
	var get_range_inter_bin_resolution = function(range_resolution, number_of_samples_per_chirp, num_range_fft_bins) {
	    return range_resolution*number_of_samples_per_chirp/num_range_fft_bins;
	};
	var get_velocity_inter_bin_resolution = function(velocity_resolution, min_num_of_chirp_loops, num_doppler_fft_bins) {
	    return velocity_resolution*min_num_of_chirp_loops/num_doppler_fft_bins;
	};
	var get_radar_cube_size = function(num_range_fft_bins, min_num_of_chirp_loops, num_virtual_rx) {
	    return num_range_fft_bins*min_num_of_chirp_loops*num_virtual_rx*4/1024;
	};
    var get_frame_periodicity = function(measurement_rate){
        return 1 / measurement_rate * 1000;
    }
    var get_chirp_time = function(ramp_end_time, idle_time_minimum){
        return ramp_end_time + idle_time_minimum;
    }
    var get_bursts_in_frame = function(device){
        if (device == 'xWRL6432' || device == 'xWRL1432'){
            return 20;
        }
        else if (device != 'xWRL6432' && device != 'xWRL1432'){
            return 1;
        }
    }
    var get_burst_periodicity = function(chirp_repetition_period, max_burst_periodicity, chirps_in_burst, num_tx){
        var burst_periodicity;
        if (chirp_repetition_period*3 < max_burst_periodicity){
            burst_periodicity = (chirp_repetition_period*chirps_in_burst)/num_tx + 115;
        }
        else{
            burst_periodicity = max_burst_periodicity - 100;
        }
        return burst_periodicity;
    }
    var get_chirps_in_burst = function(min_num_of_chirp_loops, bursts_in_frame){
        var total_chirps = min_num_of_chirp_loops*2;
        return Math.floor(total_chirps/bursts_in_frame);
    }
    var get_max_burst_periodicity = function(frame_periodicity, bursts_in_frame){
        var max_burst_periodicity = (frame_periodicity/bursts_in_frame)*1000;
        return max_burst_periodicity;
    }

    
    // ================================ END: GETTER FUNCTIONS  ==============================================================  
    // ================================ START: CFG GENERATION FUNCTIONS  ====================================================
    // SDK 3 generate cfg functions /////////////
    // Translated, Missing 2
    var generate_ChannelCfg_sdk3 = function (Input, P) {
        /*
            Presently, there are two models, AWR2243 and AWR1243, which I do not have data on.
            Need to contact the automotive group to get the proper settings here.
        */
        
        if (Output.num_rx == 4) P.channelCfg3.rxChannelEn = 15;
        else if (Output.num_rx == 3) P.channelCfg3.rxChannelEn = 7;
        else if (Output.num_rx == 2) P.channelCfg3.rxChannelEn = 3;
        else if (Output.num_rx == 1) P.channelCfg3.rxChannelEn = 2;
        else P.channelCfg3.rxChannelEn = 0;
        //P.channelCfg3.txChannelEn=IF(Number_of_TX=3,7,IF(Number_of_TX=2,5,IF(Number_of_TX=1,1,0)))
        if ((Input.device == Basic_Globals.Platforms.IWR14xx) || // IWR1443
            (Input.device == Basic_Globals.Platforms.AWR1843) || // AWR1843
            (Input.device == Basic_Globals.Platforms.IWR6843) || // IWR6843
            (Input.device == Basic_Globals.Platforms.AWR14xx))   // AWR1443
        {
            if (Input.Number_of_TX == 3) P.channelCfg3.txChannelEn = 7;
            else if (Output.num_tx == 2) P.channelCfg3.txChannelEn = 5;
            else if (Output.num_tx == 1) P.channelCfg3.txChannelEn = 1;
            else P.channelCfg3.txChannelEn = 0;
        } 
        else if (Input.device == Basic_Globals.Platforms.xWR16xx) { // xWR1642
            if (Output.num_tx == 2) P.channelCfg3.txChannelEn = 3;
            else if (Output.num_tx == 1) P.channelCfg3.txChannelEn = 1;
            else P.channelCfg3.txChannelEn = 0;
        } else {
            // The AWR2243 and AWR1243 devices will be here
            P.channelCfg3.txChannelEn = 0;
        }

        P.channelCfg3.cascading = 0;
        P.lines.push(['channelCfg', P.channelCfg3.rxChannelEn, P.channelCfg3.txChannelEn, P.channelCfg3.cascading].join(' '));
    };
    
    // Translated, Missing 2
    var generate_adcCfg = function (Output, P) {
        P.adcCfg.numADCBits = 2; //=IF(ADC_bits=16,2,"NA")
        P.adcCfg.adcOutputFmt = Output.ADC_samples_type == 2 ? 1 : 0; //=IF(ADC_samples_type=2,1,0)
        P.adcCfg.justification = 0;//TODO remove
        P.lines.push(['adcCfg', P.adcCfg.numADCBits, P.adcCfg.adcOutputFmt].join(' '));
    };
    
    // Translated, Missing 2
    var generate_adcbufCfg = function (Output, P) {
        P.dataFmt.rxChannelEn = P.channelCfg3.rxChannelEn;
        P.dataFmt.adcOutputFmt = Output.ADC_samples_type == 2 ? 0 : 1;//=IF(ADC_samples_type=2,0,1)
        P.dataFmt.chirpThreshold = Output.chirps_per_interrupt;
        P.dataFmt.SampleSwap = 1;
        
        P.dataFmt.ChanInterleave = 1;
        P.lines.push(['adcbufCfg -1', P.dataFmt.adcOutputFmt, P.dataFmt.SampleSwap, P.dataFmt.ChanInterleave, P.dataFmt.chirpThreshold].join(' '));
    };
    
    // Translated, Missing 2
    var generate_profileCfg = function (Output, P) {
        P.lines.push(['profileCfg 0', Output.starting_frequency, Output.idle_time_minimum, Output.adc_valid_start_time_2, 
            Util.Format.toPrecision(Output.ramp_end_time, 2), Util.Format.toPrecision(Output.ramp_slope, 2), Output.number_of_samples_per_chirp, 
            Util.Format.toPrecision(Output.sampling_frequency * 1000, 0), '0 0 30'].join(' '));
    };

    // Translated, Missing 2
    var generate_chirpCfg = function (Input, P) {
        /* This is the implementation used in the Demo Visualizer: */
        var chirpCfg = {}; P.chirpCfg.push(chirpCfg);
        
        chirpCfg.startIdx = 0;
        chirpCfg.endIdx = 0;
        chirpCfg.profileId = 0;
        chirpCfg.startFreq = 0;
        chirpCfg.freqSlopeVar = 0;
        chirpCfg.idleTime = 0;
        chirpCfg.adcStartTime = 0;
        //chirpCfg.txEnable = 1;
        if ((Input.device == Basic_Globals.Platforms.AWR1843) ||
            (Input.device == Basic_Globals.Platforms.IWR6843))  
        {
            if (Output.num_tx == 3) chirpCfg.txEnable = 1;
            else if (Output.num_tx == 2) chirpCfg.txEnable = 1;
            else chirpCfg.txEnable = 1;
        } 
        else if (Input.device == Basic_Globals.Platforms.xWR16xx) 
        {
            if (Output.num_tx == 2) chirpCfg.txEnable = 1;
            else chirpCfg.txEnable = 1;
        } 
        else 
        {
            chirpCfg.txEnable = 0;
        }
        
        chirpCfg = {}; P.chirpCfg.push(chirpCfg);
        chirpCfg.startIdx = 1;
        chirpCfg.endIdx = 1;
        chirpCfg.profileId = 0;
        chirpCfg.startFreq = 0;
        chirpCfg.freqSlopeVar = 0;
        chirpCfg.idleTime = 0;
        chirpCfg.adcStartTime = 0;
        if ((Input.device == Basic_Globals.Platforms.AWR1843) ||
            (Input.device == Basic_Globals.Platforms.IWR6843)) 
        {
            if (Output.num_tx == 4) chirpCfg.txEnable = 4;
            else if (Output.num_tx == 3) chirpCfg.txEnable = 4;
            else if (Output.num_tx == 2) chirpCfg.txEnable = 4;
            else chirpCfg.txEnable = 0;
        } else if (Input.device == Basic_Globals.Platforms.xWR16xx) {
            if (Output.num_tx == 2) chirpCfg.txEnable = 2;
            else chirpCfg.txEnable = 0;
        } else {
            chirpCfg.txEnable = 0;
        }
        
        if (((Input.device == Basic_Globals.Platforms.xWR18xx) ||
             (Input.device == Basic_Globals.Platforms.AM273X)   ||
             (Input.device == Basic_Globals.Platforms.AWR294X) || //TODO
             (Input.device == Basic_Globals.Platforms.xWR64xx) ||
             (Input.device == Basic_Globals.Platforms.xWR68xx)) && Input.Number_of_TX >= 3) { // TODO 3D case
            chirpCfg = {}; P.chirpCfg.push(chirpCfg);
            chirpCfg.startIdx = 2;
            chirpCfg.endIdx = 2;
            chirpCfg.profileId = 0;
            chirpCfg.startFreq = 0;
            chirpCfg.freqSlopeVar = 0;
            chirpCfg.idleTime = 0;
            chirpCfg.adcStartTime = 0;
            if (Input.device == Basic_Globals.Platforms.AWR294X){
                if(Input.Number_of_AzimTX == 3){
                    chirpCfg.txEnable = 8; // 3 Azim
                }
                else{
                    chirpCfg.txEnable = 2; // 2 Azim, 1 Elev
                }
            }
            else{
                chirpCfg.txEnable = 2;
            }
        } 
        if(Input.device == Basic_Globals.Platforms.AWR294X && Input.Number_of_TX == 4){
            chirpCfg = {}; P.chirpCfg.push(chirpCfg);
            chirpCfg.startIdx = 3;
            chirpCfg.endIdx = 3;
            chirpCfg.profileId = 0;
            chirpCfg.startFreq = 0;
            chirpCfg.freqSlopeVar = 0;
            chirpCfg.idleTime = 0;
            chirpCfg.adcStartTime = 0;
            chirpCfg.txEnable = 2; 
        }
        for (var idx = 0; idx < P.chirpCfg.length; idx++) {
            chirpCfg = P.chirpCfg[idx];
            P.lines.push(['chirpCfg', chirpCfg.startIdx, chirpCfg.endIdx, chirpCfg.profileId, chirpCfg.startFreq, chirpCfg.freqSlopeVar,
                chirpCfg.idleTime, chirpCfg.adcStartTime, chirpCfg.txEnable].join(' '));
        }
    };
    
    var generate_frameCfg3 = function (Input, Output, P) {
        P.lines.push(['frameCfg 0 0',Output.min_num_of_chirp_loops,'0',Util.Format.toPrecision(1 / Input.measurement_rate * 1000, 2),'1 0'].join(' '));
    };
    
    // SDK 5 generate cfg functions /////////////
    var generate_ChannelCfg_sdk5 = function (Input, P) {
        /*
            Presently, there are two models, AWR2243 and AWR1243, which I do not have data on.
            Need to contact the automotive group to get the proper settings here.
            
            Configuration support for xWRL6432 coming soon
        */
        
        if (Output.num_rx == 3) P.channelCfg5.rxChCtrlBitMask = 7;
        else if (Output.num_rx == 2) P.channelCfg5.rxChCtrlBitMask = 3;
        else if (Output.num_rx == 1) P.channelCfg5.rxChCtrlBitMask = 2;
        else P.channelCfg.rxChCtrlBitMask = 0;
        //P.channelCfg5.txChCtrlBitMask=IF(Number_of_TX=2,5,IF(Number_of_TX=1,1,0))
        if (Input.device == Basic_Globals.Platforms.xWRL6432 || Input.device == Basic_Globals.Platforms.xWRL1432) // xWRL6432
        {
            if (Output.num_tx == 2) P.channelCfg5.txChCtrlBitMask = 3;
            else if (Output.num_tx == 1) P.channelCfg5.txChCtrlBitMask = 1;
            else P.channelCfg5.txChCtrlBitMask = 0;
        } else {
            // Future SDK 5 devices with different antenna numbers can be here
            P.channelCfg5.txChCtrlBitMask = 0;
        }

        P.channelCfg.miscCtrl = 0;
        P.lines.push(['channelCfg', P.channelCfg5.rxChCtrlBitMask, P.channelCfg5.txChCtrlBitMask, P.channelCfg5.miscCtrl].join(' '));
    };

    var generate_chirpComnCfg = function (Output, P) {
        /* These are the outputs needed
        DigOutputSampRate
        DigOutputBitsSel
        DfeFirSel
        NumAdcSamples
        ChirpTxMimoPatSel
        ChirpRampEndTime
        ChirpRxHpfSel
        */
        
        P.lines.push(['chirpComnCfg',Util.Format.toPrecision(100/Output.sampling_frequency, 0),'0','0',
            Output.number_of_samples_per_chirp, '4', Util.Format.toPrecision(Output.ramp_end_time, 0),'chirprxhpfsel'].join(' '));
        
       /* profile config P.lines.push(['profileCfg 0', Output.starting_frequency, Output.idle_time_minimum, Output.adc_valid_start_time_2, 
            Util.Format.toPrecision(Output.ramp_end_time, 2), Util.Format.toPrecision(Output.ramp_slope, 2), Output.number_of_samples_per_chirp, 
           Util.Format.toPrecision(Output.sampling_frequency * 1000, 0), '0 0 30'].join(' ')); */
    };

    var generate_chirpTimingCfg = function (Output, P) {
        /* These are the outputs needed
            ChirpIdleTime
            ChirpAdcSkipSamples
            ChirpTxStartTime
            ChirpRfFreqSlope
            ChirpRfFreqStart
        */
        
        P.lines.push(['chirpTimingCfg ',Output.idle_time_minimum, 'chirp skip samples','0',Util.Format.toPrecision(Output.ramp_slope, 2),Output.starting_frequency]);
        
       /* profile config P.lines.push(['profileCfg 0', Output.starting_frequency, Output.idle_time_minimum, Output.adc_valid_start_time_2, 
            Util.Format.toPrecision(Output.ramp_end_time, 2), Util.Format.toPrecision(Output.ramp_slope, 2), Output.number_of_samples_per_chirp, 
           Util.Format.toPrecision(Output.sampling_frequency * 1000, 0), '0 0 30'].join(' ')); */
    };
    
    var generate_frameCfg5 = function (Input, Output, P){
        /* These are the outputs needed
        NumOfChirpsInBurst = numLoops
        NumOfChirpsAccum = 0
        BurstPeriodicity
        NumOfBurstsInFrame = 1
        FramePeriodicity = Util.Format.toPrecision(1 / Input.measurement_rate * 1000, 2) in ms
        NumOfFrames = 0
        */

        P.lines.push(['frameCfg ',Output.chirps_in_burst,' 0',Util.Format.toPrecision(Output.burst_periodicity), Output.bursts_in_frame ,Util.Format.toPrecision(1 / Input.measurement_rate * 1000, 2),' 0']);
        
    }

    // Generate cfg output /////////////////////
    
    var generateCfg = function () {
        var Input = this.Input;
        var Output = this.Output;
        
        if (Input.device != Basic_Globals.Platforms.xWRL6432 && Input.device != Basic_Globals.Platforms.xWRL1432){
            var P = {
                channelCfg3: {}, adcCfg: {}, dataFmt: {}, profileCfg: {},
                chirpCfg: [], frameCfg3: {}, guiMonitor: {}, clutterRemoval: {}, lines: []
            };
        }
        else if(Input.device == Basic_Globals.Platforms.xWRL6432 || Input.device == Basic_Globals.Platforms.xWRL1432){
            var P = {
                channelCfg5: {}, chirpComnCfg:{}, chirpTimingCfg:{}, frameCfg5: {}, lines: []
            };
        }
        
        // Error check on the models that I do not currently have data for:
        if (Input.device == Basic_Globals.Platforms.AWR1243 || Input.device == Basic_Globals.Platforms.AWR2243) {
            P.lines.push('% ***************************************************************');
            P.lines.push([' FRONT END DEVICE ONLY, DOES NOT REQUIRE CONFIGURATION FILE:', Input.device].join(' '));
            P.lines.push('% ***************************************************************');
            return P;
        }
        
        // Otherwise, I can actually generate the CFG file
        P.lines.push('% ***************************************************************');
        //P.lines.push(['% Verified for SDK:3.5 and ealier'); // Add this line in once we verify that these cmds work for the 60Lo device
        P.lines.push(['% Created using Visualizer ver', visualizerVersion].join(': '));
        P.lines.push(['% Basic_Globals.Platforms', Input.device].join(': '));
        P.lines.push(['% Number of Rx Antennas', Output.num_rx].join(': '));
        P.lines.push(['% Number of Tx Antennas', Output.num_tx].join(': '));
        P.lines.push(['% Transmit Antenna Gain (dB)', Output.tx_gain].join(': '));
        P.lines.push(['% Receive Antenna Gain (dB)', Output.rx_gain].join(': '));
        P.lines.push(['% Frequency', Input.frequency_range].join(': '));
        P.lines.push(['% Maximum Bandwidth (MHz)', Output.maximum_bandwidth].join(': '));
        P.lines.push(['% Transmit Power (dBm)', Output.tx_power].join(': '));
        P.lines.push(['% Maximum Detectable Range (m)', Input.maximum_detectable_range].join(': '));
        P.lines.push(['% Range Resolution (cm)', Input.range_resolution].join(': '));
        P.lines.push(['% Maximum Velocity (km/h)', Input.maximum_velocity_kmph].join(': '));
        P.lines.push(['% Velocity Resolution (km/h)', Input.velocity_resolution_kmph].join(': '));
        P.lines.push(['% Measurement Rate (Hz)', Input.measurement_rate].join(': '));
        P.lines.push(['% Typical Detectable Object (m^2)', Input.typical_detected_object].join(': '));
        P.lines.push(['% Detection Loss (dB)', Input.detection_loss].join(': '));
        P.lines.push(['% System Loss (dB)', Input.system_loss].join(': '));
        P.lines.push(['% Implementaion Margin (dB)', Input.implementation_margin].join(': '));
        P.lines.push(['% Detection SNR (dB)', Input.detection_SNR].join(': '));
        P.lines.push('% ***************************************************************');

        P.lines.push('sensorStop');
        P.lines.push('flushCfg');
        P.lines.push('dfeDataOutputMode 1');

        if (Input.device != Basic_Globals.Platforms.xWRL6432 && Input.device != Basic_Globals.Platforms.xWRL1432){
            // channelCfg sdk3
            generate_ChannelCfg_sdk3(Input, P);

            // adcCfg
            generate_adcCfg(Output, P);

            // dataFmt (adcbufCfg)
            generate_adcbufCfg(Output, P);

            // profileCfg
            generate_profileCfg(Output, P);

            // chirpCfg 
            generate_chirpCfg(Input, P);

            // frameCfg
            generate_frameCfg3(Input, Output, P);
        }
        else if(Input.device == Basic_Globals.Platforms.xWRL6432 || Input.device == Basic_Globals.Platforms.xWRL1432){
            // channelCfg sdk5
            generate_ChannelCfg_sdk5(Input, P);

            //chirpComnCfg
            generate_chirpComnCfg(Output, P);

            //chirpTimingCfg
            generate_chirpTimingCfg(Output, P);

            //chirpTimingCfg
            generate_frameCfg5(Input, Output, P);
        }
        /*

        // lowPower
        generate_lowPowerCfg(Input, P);

        // guiMonitor
        generate_guiMonitorCfg(Input, P);

        // cfarCfg, cfarRangeCfg, cfarDopplerCfg
        generate_cfarCfg(Input, P);

        // peakGrouping
        generate_peakGroupingCfg(Input, P);

        // multiObjBeamForming
        generate_BFCfg(Input, P);

        //Static clutter removal
        generate_clutterCfg(Input, P);

        //always disable calibDcRangeSig and fill in defaults (needed as they are checked by demo)
        generate_DcRangeCfg(Input, P);

        //always disable extendedMaxVelocity
        generate_extendedVeloCfg(Input, P);

        //always disable BPM
        generate_bpmCfg(Input, P);

        //always disable LVDS streaming
        generate_lvdsStreamCfg(Input, P);

        //always disable nearField
        generate_nearFieldCfg(Input, P);

        // compensate range and angle based on input
        generate_compRangeBiasAndRxChanPhase(Input, P);

        // disable always
        generate_measureRangeBiasAndRxChanPhase(Input, P);

        // CQ rxSaturation
        generate_CQrxSat(Input, P);

        // CQ Signal Image band monitor
        generate_CQSigImg(Input, P);

        // Analog monitor
        generate_analogMon(Input, P);


        */

        P.lines.push('sensorStart');
        return P;
    };
    // ================================ END: CFG GENERATION FUNCTIONS  ======================================================
    
    // ================================ START: OBJECT EXPORTATION  ==========================================================
    mmWaveBasic.prototype.init = init;
    mmWaveBasic.prototype.updateInput = updateInput;
    mmWaveBasic.prototype.generateCfg = generateCfg;
    mmWaveBasic.prototype.get_num_rx = get_num_rx;
    mmWaveBasic.prototype.get_num_tx = get_num_tx;
    mmWaveBasic.prototype.get_evm_list = get_evm_list;
    mmWaveBasic.prototype.get_allowed_freq_range = get_allowed_freq_range;
    mmWaveBasic.prototype.get_typical_detected_object_to_rcs = get_typical_detected_object_to_rcs;

    // export as AMD/CommonJS module or global variable
    if (typeof define === 'function' && define.amd) define('mmWaveBasic', function () { return mmWaveBasic; });
    else if (typeof module !== 'undefined') module.exports = mmWaveBasic;
    else if (typeof self !== 'undefined') self.mmWaveBasic = mmWaveBasic;
    else window.mmWaveBasic = mmWaveBasic;
    // ================================ END: OBJECT EXPORTATION  ============================================================
})();
