// jshint maxerr:1000
/*
 * Copyright (c) 2023, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* 
 * INDEX.JS ------------------------------------------------------------------------------------
 * This file serves as the central/main file for the application. 
 * It implements the bare necessities, the intantiates the necessary classes and objects.
 * -------------------------------------------------------------------------------------------*/

// -------------------------------------------- GLOBALS ---------------------------------------------------------------
var initComplete = false;   // Used for the init function to check if the template has been stamped
var templateObj;            // Used to access the application template

// -------------------------------------------- CLASS OBJECTS ---------------------------------------------------------
var Util;
var mmWaveBasic;
var mmWaveAdvanced;
var mmWavePower;

// -------------------------------------------- EVENT LISTENERS -------------------------------------------------------
/**
 *  This code registers a actions with the GUI Composer navigation bar. You can add new ones here, or modify the existing ones.
 *  @example function gc.nav.registryAction(id, runable, [isAvailable], [isVisible]);
 *
 *  @param {*} id - uniquely identifies the action, and should correspond to the action property of the menuaction widget.
 *  @param {*} runable - function that performs the custom action.
 *  @param {*} isAvailable - (optional) - function called when the menu action is about to appear.  Return false to disable the action, or true to enable it.
 *  @param {*} isVisible - (optional) - function called when the menu action is about to appear.  Return false to hide the action, or true to make it visible.
 */
document.addEventListener('gc-nav-ready', function() {
    gc.nav.registerAction('HelpAbout', {
        run: function() {
            templateObj.$.dialog_help_about.open();
        },
        isAvailable: function() {
            return true;
        },
        isVisible: function() {
            return true;
        }
    });
    gc.nav.registerAction('ti_widget_menuaction_userguide', {
        run: function() {
            templateObj.$.userguideViewer.open();
        },
        isAvailable: function() {
            return true;
        },
        isVisible: function() {
            return true;
        }
    });
    gc.nav.registerAction('ti_widget_menuaction_download', {
        run: function() {
            window.open('https://dev.ti.com/gallery/info/1792614/mmWaveSensingEstimator/', '_blank');
        },
        isAvailable: function() {
            return true;
        },
        isVisible: function() {
            return true;
        }
    });
});

/**
 *  Create a global onChange callback listener to catch user input.
 * 
 *  1. The window.addEventListener() method attaches an event handler to the document.
 *  2. The change event occurs when the value of an <input>, <select>, or <textarea> element has been changed.
 *  3. The for loop iterates through the html event element path to see where in the tool change/input has occurred,
 *     then we act accordingly. 
*/
window.addEventListener('change', function(event) {
    // Iterate through the html element path to see if the change occurred on the first tab or the second:
    for (var i in event.composedPath()) {
        // Ignore user input in the case of 
        //  - ti_widget_droplist_desired_cfg
        //  - ti_widget_droplist_device 
        // as it is specifically handled in on-change functions of their own, bound to the respective html element.
        if(event.composedPath()[i].id == 'ti_widget_droplist_desired_cfg' || 
           event.composedPath()[i].id == 'ti_widget_droplist_device')
        {
            break;
        }
        // Basic Tool Input
        else if (event.composedPath()[i].id == 'ti_widget_container_input_basic') {
            Basic_onInput();
            break;
        }
        // Advanced Tool Input
        // -----> Upper Section called "Configuration Inputs"
        else if (event.composedPath()[i].id == 'ti_widget_container_cc_inputs_advanced') {
            Adv_onInput();
            break;
        }
        // -----> Lower Section called "Chirp Parameters"
        else if (event.composedPath()[i].id == 'ti_widget_container_cc_outputs_advanced') {
            Adv_onIntermediateInput(); // Take a change in the Configurations Parameter section, copy it to the inputs, then run normally 
            break;
        }
        // Power Tool Input
        else if (event.composedPath()[i].id == 'ti_widget_container_input_power') {
            Power_onInput();
            break;
        }
    }
    saveObjCookie();
}); 

// -------------------------------------------- INITIALIZATION --------------------------------------------------------
/**
 *  Expose the template object to the global scope so that it can be accessed from our code.
 *  Also, call other one-time initialization functions.
 */
var init = function() {
    templateObj = document.querySelector('#template_obj');
    
    // Wait for DOMContentLoaded event before trying to access the application template:
    // template to fires a dom-change event to indicate that it has been 'stamped'
    // before trying to access components in the application.
    templateObj.addEventListener('dom-change',function() {
        if (initComplete) return;
        this.async(function() {
            // Stamp application, then you can use 'automatic node finding' $ syntax to access widgets: templateObj.$.widgetId
            initComplete = true;

            Util            = new Util();               // Create the backend for the util functions;
            mmWaveBasic     = new mmWaveBasic();        // Create the backend for the basic tab tool
            mmWaveAdvanced  = new mmWaveAdvanced();     // Create the backend for the advanced tab tool
            mmWavePower     = new mmWavePower();        // Create the backend for the power tab tool

            // Add Favicon to the browser
            Util.Html.addFavicon();

            // TODO Remove with the reintegration of cookies:
            saveObjCookie(); // Actually deletes cookies

            // Based on the default dropdown, fill out default values:
            loadObjCookie();
        },1);
    });
};

/**
 * If the template tag was found then call the init function, otherwise wait for the DOM to be loaded
 */
var templateObj = document.querySelector('#template_obj');
if (templateObj) {
    init();
} else {
    document.addEventListener('DOMContentLoaded',init.bind(this));
}

// -------------------------------------------- FUNCTIONS ------------------------------------------------------------- 
/**
 * Save the current state of the tool to a cookie TODO: reimplement this function
 */
var saveObjCookie = function() {
    // TODO: Cookies are disabled for now, but we may want to enable them later
    //bakeCookie('basicInput',mmWaveBasic.Input);     // TODO ENABLE COOKIES LATER
    //bakeCookie('advInput',mmWaveAdvanced.Input);    // TODO ENABLE COOKIES LATER
    //bakeCookie('powerInput',mmWavePower.Input);    // TODO ENABLE COOKIES LATER
    // For now, delete cookies that people may have
    Util.Cookie.deleteCookie('basicInput');
    Util.Cookie.deleteCookie('advInput');
    Util.Cookie.deleteCookie('powerInput');
};

/**
 * Load the current state of the tool to a cookie TODO: reimplement this function
 */          
var loadObjCookie = function() {
    // TODO, need to reimplement this function to work with our new file division
    // Load the object from cookies:
    var res = Util.Cookie.loadCookie('basicInput');
    if (res !== null) {
        mmWaveBasic.Input = res;
        Basic_onInput('init');
        
    } else {
        Basic_onDesiredCfg();
    }

    res = Util.Cookie.loadCookie('advInput');
    if (res !== null) {
        mmWaveAdvanced.Input = res;
        Adv_onInput('init');
    } else {
        Adv_onInput();
    }
    
    res = Util.Cookie.loadCookie('powerInput');
    if (res !== null) {
        mmWavePower.Input = res;
        Power_onInput('init');
    } else {
        Power_onInput();
    }
};